@@ -15,12 +15,11 @@
 # You should have received a copy of the GNU General Public License
 # along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
 
-# Make coding more python3-ish
-from __future__ import (absolute_import, division, print_function)
-__metaclass__ = type
+from __future__ import annotations
 
 import ast
 import datetime
+import functools
 import os
 import pwd
 import re
@@ -49,7 +48,7 @@ from ansible.module_utils.six import string_types
 from ansible.module_utils.common.text.converters import to_native, to_text, to_bytes
 from ansible.module_utils.common.collections import is_sequence
 from ansible.plugins.loader import filter_loader, lookup_loader, test_loader
-from ansible.template.native_helpers import AnsibleUndefined, ansible_native_concat, ansible_eval_concat, ansible_concat
+from ansible.template.native_helpers import ansible_native_concat, ansible_eval_concat, ansible_concat
 from ansible.template.template import AnsibleJ2Template
 from ansible.template.vars import AnsibleJ2Vars
 from ansible.utils.display import Display
@@ -297,23 +296,7 @@ def _unroll_iterator(func):
             return list(ret)
         return ret
 
-    return _update_wrapper(wrapper, func)
-
-
-def _update_wrapper(wrapper, func):
-    # This code is duplicated from ``functools.update_wrapper`` from Py3.7.
-    # ``functools.update_wrapper`` was failing when the func was ``functools.partial``
-    for attr in ('__module__', '__name__', '__qualname__', '__doc__', '__annotations__'):
-        try:
-            value = getattr(func, attr)
-        except AttributeError:
-            pass
-        else:
-            setattr(wrapper, attr, value)
-    for attr in ('__dict__',):
-        getattr(wrapper, attr).update(getattr(func, attr, {}))
-    wrapper.__wrapped__ = func
-    return wrapper
+    return functools.update_wrapper(wrapper, func)
 
 
 def _wrap_native_text(func):
@@ -326,7 +309,36 @@ def _wrap_native_text(func):
         ret = func(*args, **kwargs)
         return NativeJinjaText(ret)
 
-    return _update_wrapper(wrapper, func)
+    return functools.update_wrapper(wrapper, func)
+
+
+class AnsibleUndefined(StrictUndefined):
+    '''
+    A custom Undefined class, which returns further Undefined objects on access,
+    rather than throwing an exception.
+    '''
+    def __getattr__(self, name):
+        if name == '__UNSAFE__':
+            # AnsibleUndefined should never be assumed to be unsafe
+            # This prevents ``hasattr(val, '__UNSAFE__')`` from evaluating to ``True``
+            raise AttributeError(name)
+        # Return original Undefined object to preserve the first failure context
+        return self
+
+    def __getitem__(self, key):
+        # Return original Undefined object to preserve the first failure context
+        return self
+
+    def __repr__(self):
+        return 'AnsibleUndefined(hint={0!r}, obj={1!r}, name={2!r})'.format(
+            self._undefined_hint,
+            self._undefined_obj,
+            self._undefined_name
+        )
+
+    def __contains__(self, item):
+        # Return original Undefined object to preserve the first failure context
+        return self
 
 
 class AnsibleContext(Context):
@@ -464,11 +476,11 @@ class JinjaPluginIntercept(MutableMapping):
             self._seen_it.remove(key)
             raise TemplateSyntaxError('Could not load "%s": %s' % (key, to_native(original_exc or e)), 0)
 
-        # if i do have func and it is a filter, it nees wrapping
+        # if i do have func and it is a filter, it needs wrapping
         if self._pluginloader.type == 'filter':
             # filter need wrapping
             if key in C.STRING_TYPE_FILTERS:
-                # avoid litera_eval when you WANT strings
+                # avoid literal_eval when you WANT strings
                 func = _wrap_native_text(func)
             else:
                 # conditionally unroll iterators/generators to avoid having to use `|list` after every filter
@@ -694,7 +706,7 @@ class Templar:
             setattr(obj, key, original[key])
 
     def template(self, variable, convert_bare=False, preserve_trailing_newlines=True, escape_backslashes=True, fail_on_undefined=None, overrides=None,
-                 convert_data=True, static_vars=None, cache=None, disable_lookups=False):
+                 convert_data=True, static_vars=None, disable_lookups=False):
         '''
         Templates (possibly recursively) any given data as input. If convert_bare is
         set to True, the given data will be wrapped as a jinja2 variable ('{{foo}}')
@@ -702,9 +714,6 @@ class Templar:
         '''
         static_vars = [] if static_vars is None else static_vars
 
-        if cache is not None:
-            display.deprecated("The `cache` option to `Templar.template` is no longer functional, and will be removed in a future release.", version='2.18')
-
         # Don't template unsafe variables, just return them.
         if hasattr(variable, '__UNSAFE__'):
             return variable
@@ -871,11 +880,9 @@ class Templar:
             return [] if wantlist else None
 
         if not is_sequence(ran):
-            display.deprecated(
+            raise AnsibleLookupError(
                 f'The lookup plugin \'{name}\' was expected to return a list, got \'{type(ran)}\' instead. '
                 f'The lookup plugin \'{name}\' needs to be changed to return a list. '
-                'This will be an error in Ansible 2.18',
-                version='2.18'
             )
 
         if ran and allow_unsafe is False:
@@ -1010,12 +1017,16 @@ class Templar:
                     if unsafe:
                         res = wrap_var(res)
             return res
-        except (UndefinedError, AnsibleUndefinedVariable) as e:
+        except UndefinedError as e:
             if fail_on_undefined:
-                raise AnsibleUndefinedVariable(e, orig_exc=e)
-            else:
-                display.debug("Ignoring undefined failure: %s" % to_text(e))
-                return data
+                raise AnsibleUndefinedVariable(e)
+            display.debug("Ignoring undefined failure: %s" % to_text(e))
+            return data
+        except AnsibleUndefinedVariable as e:
+            if fail_on_undefined:
+                raise
+            display.debug("Ignoring undefined failure: %s" % to_text(e))
+            return data
 
     # for backwards compatibility in case anyone is using old private method directly
     _do_template = do_template
